/***************
 * NAME:                  "ViewTIFF" - viewtiff.c
 * DESCRIPTION:           Functions for loading tiff images
 * AUTHORS:               Mike Riddle 
 * DATE                   03.11.21
 * RELEASE                initial release version 1.0.0
 ***************/
 /* Revised 03.12.21 - version 1.0.1 - Added support to Load TIFF. */
 /* Revised 03.15.21 - version 1.0.2 - Added Menu. Added support to Save PNG. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <intuition/intuition.h>
#include <cybergraphx/cybergraphics.h>
#include <proto/exec.h>
#include <proto/asl.h>
#include <proto/dos.h>
#include <proto/icon.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <proto/gadtools.h>
#include <proto/datatypes.h>
#include <datatypes/pictureclass.h>
#include <proto/cybergraphics.h>

#include <tiff.h>
#include <tiffio.h>

#define NO (0)
#define YES (1)

/**************************************************************************************************/

struct SDT_picture {
UWORD width,height,bits,padding;
ULONG format;
UBYTE *pixels;
};

/****************************************************************************************/

struct IconInfo {
	ULONG Width;
	ULONG  Height;
	ULONG  Format;
	ULONG  NumColors;
	const struct ColorRegister *CMap;
	UBYTE *Image;
	ULONG  Which;
	ULONG  NumImages;
};

/****************************************************************************************/

typedef struct
{
    UWORD Width;
	UWORD Height;
	UWORD Depth;
	UBYTE *CMAP;
    ULONG *Cregs;
    struct ColorRegister *Cmap;
	UBYTE *Data;
}BitMapImage;

/****************************************************************************************/

typedef struct
{
	UWORD Width;
	UWORD Height;
	char *Type;
	UBYTE *Data;
}RGBImage;

/****************************************************************************************/

/* Message goes in the %s parameter */
   struct EasyStruct ResultReq = {
     sizeof(struct EasyStruct),
     0,
     "Message",
     "%s",
     "OK"
   };

/****************************************************************************************/

/* Message goes in the %s parameter */
   struct EasyStruct YesNoReq = {
     sizeof(struct EasyStruct),
     0,
     "Question",
     "%s",
     "Yes|No"
   };

/****************************************************************************************/

typedef struct
//struct BitMapHeader
{
    UWORD        bmh_Width;
    UWORD        bmh_Height;
    WORD         bmh_Left;
    WORD         bmh_Top;
    UBYTE        bmh_Depth;
    UBYTE        bmh_Masking;
    UBYTE        bmh_Compression;
    UBYTE        bmh_Pad;
    UWORD        bmh_Transparent;
    UBYTE        bmh_XAspect;
    UBYTE        bmh_YAspect;
    WORD         bmh_PageWidth;
    WORD         bmh_PageHeight;
}__attribute__((packed)) BitMapHeader;

/****************************************************************************************/

                              //** MAIN MENU ITEMS **//

/****************************************************************************************/

struct NewMenu newmenu[] = {

	{NM_TITLE,"Project",            NULL,0,0,0}, //menu 0
	{ NM_ITEM,"File...",             "S" ,0,0,0},	//0
	{ NM_ITEM,NM_BARLABEL,          NULL,0,0,0}, //1    
	{ NM_ITEM,"Save As...",             0,0,0,0}, //2
	{ NM_ITEM,"Save As TIFF",            0,0,0,0}, //3	
	{ NM_ITEM,"Save As PNG",            0,0,0,0}, //4
	{ NM_ITEM,NM_BARLABEL,          NULL,0,0,0}, //5
	{ NM_ITEM,"About",                "A" ,0,0,0}, //6
	{ NM_ITEM,NM_BARLABEL,          NULL,0,0,0}, //7
	//{ NM_ITEM,"Help...",                "A" ,0,0,0},
	//{ NM_ITEM,NM_BARLABEL,          NULL,0,0,0},
	{ NM_ITEM,"Exit...",                0,0,0,0}, //8
	
	{NM_TITLE,"Image...",            NULL,0,0,0}, //menu 1
	{ NM_ITEM,"TIFF Data",             "S" ,0,0,0},
	{ NM_ITEM,"TIFF Tags",             "S" ,0,0,0},
    
    {  NM_END}
	};

/****************************************************************************************/

                        //** MAIN FUNCTION DECLARATIONS **//

/****************************************************************************************/

BOOL EXISTS(char *fname);
char *get_ext(char* path);
BOOL Load_TIFF(char *file_name); 
BOOL LoadAsDatatype(char *file_name);    
BOOL testImage_by_datatype(char *file_name);
//ULONG DoWriteMethod(Object *dto, STRPTR name, ULONG mode);
ULONG DoWriteMethod(Object *dto, STRPTR name, struct Window *win, ULONG mode);
int CMAP2Object(Object *o, UBYTE *rgb, ULONG rgbsize );
char *DetectImageType_by_datatype(char *file_name);
struct ColorRegister *Cmap_by_datatype(char* fname);
STRPTR GetFileName(char **filepart, char *titletext);
STRPTR GetNewFileName(char **filepart, char *titletext);
struct BitMapHeader *readbmhd_by_datatype(char *file_name);
BOOL SaveAsDatatype(char *old_fname, char *filename, char *pType);
BOOL readpic_by_datatype(RGBImage *pic, char *file_name);
BOOL readpic_by_datatype_alpha(RGBImage *pic, char *file_name);
BOOL readpic_by_datatype_lut8(BitMapImage *bmi, char *file_name);
BOOL savepic_by_datatype(RGBImage *pic, char *file_name, char *type);
BOOL savepic_by_datatype_alpha(RGBImage *pic, char *file_name, char *type);
BOOL savepic_by_datatype_lut8(BitMapImage *bmi, UBYTE *colorBytes, char *file_name, char *type);
void WriteFile(char *outFileName, char *data, LONG offset, LONG fileSize);
int showImage(char *bitmapData, int width, int height, int depth, char *fpath);
BOOL Save_TIFF(char *output_file, unsigned char *image, int wd, int ht, int dp);
BOOL savepix_by_datatype(unsigned char *image, char *fileToSave, int w, int h, char *type);
BOOL savepix_by_datatype_alpha(unsigned char *image, char *fileToSave, int w, int h, char *type);
void ConvertRGBAImage2RGB(unsigned char *raster, unsigned char *RGBImage, LONG wd, LONG ht);
void ConvertRGBAImage2ARGB(unsigned char *raster, unsigned char *ARGBImage, LONG wd, LONG ht, char *PixelFormat);
void Flip_Vertical(unsigned char *pixels_buffer, LONG width, LONG height, int bytes_per_pixel);

/****************************************************************************************/

                        //** MAIN PROGRAM STARTS HERE **//

/****************************************************************************************/

char *name;
int nameCount, winCount;
char *fileNameBuffer[50];

int main(int argc, char *argv[])
{


//int rc = RETURN_FAIL;
//struct Screen *scr; Object *dto;
//struct BitMapHeader *bmhd;
//struct BitMap *bm;
//char *name;
//WORD winw,winh;
//struct Window *win;
//struct IntuiMessage *imsg;
//BOOL cont;

//name = argv[1];
//LONG dataOffset;
//LONG dataLength;
//LONG fileLength;
//LONG c1, r1, w1 = 0;
BOOL ret = FALSE; 


//if(argc == 0) [Workbench]. //if(argc < 2) [Shell].
if(argc == 0)
{
    //If no file is selected show "Usage" or use "Requester" to get filename       
    char *filePart;    
    STRPTR Image = GetFileName(&filePart, "Open File");
        
    //If nothing selected or no file selected    
    if((!Image) || (strlen(filePart) == 0))
    {
        char *msgString = "\nNo Picture File Selected!\n";
        EasyRequest(NULL, &ResultReq, NULL, msgString); 
        return 0;
    }   

    //Check for tiff files    
    char *ext = get_ext(Image);    
    if((strnicmp((const char *)ext, "tif", 4)==0) || (strnicmp((const char *)ext, "tiff", 4)==0))
    {
        /* Load Image Using Lib TIFF */
	    EasyRequest(NULL, &ResultReq, NULL, "Load TIFF Image");
	    ret = Load_TIFF(Image);	    

        /* Load Image Using Datatype */ 
		//ret = LoadAsDatatype(Image);

        if(ret == FALSE)
		    EasyRequest(NULL, &ResultReq, NULL, "Failed to Load TIFF Image!");
	
	    //ret = Save_Tiff(Image);
	    //if(ret == FALSE)
		    //EasyRequest(NULL, &ResultReq, NULL, "Failed to Save TIFF Image!");
    }
    else
    {
        /* Load Image Using Picture DataTypes */
	    EasyRequest(NULL, &ResultReq, NULL, "Load Std Image");	    
        ret = LoadAsDatatype(Image);
    }

}
else
{
    
//** Prevent users from running program in shell **//

printf("\n\n\n=====================================================");
printf("\nTHIS PROGRAM CAN'T RUN FROM AROS SHELL. USE WORKBENCH");
printf("\n=====================================================\n\n");


}

return 0;

} //End Main


/***********************************************/
//**       HELPER FUNCTIONS START HERE       **//
/***********************************************/

BOOL EXISTS(char *fname)
{
    FILE *file;
    if ((file = fopen(fname, "r")))
    {
        fclose(file);
        return TRUE;
    }
    return FALSE;
}

/**************************************************************************************************/

char* getFileNameFromPath(char* path)
{
    int i = 0;
    for(i = strlen(path) - 1;  i >= 0; i--)
    {
        //Includes Full Path with File Name
        if (path[i] == '/' && path[i] != ':')
        {
            return &path[i+1];
        }

        //Includes Volume + File Name Only!
        if (path[i] == ':' && path[i] != '/')
        {
            return &path[i+1];
        }
    }

   return path;
}

/**************************************************************************************************/

char *get_ext(char *filename)
{
	static char extension[32];
	int position=strlen((char *)filename)-1;

	strcpy(extension,"");

	while(position > -1 && filename[position] != '.') position--;

	if (position > -1)
	{
		strncpy(extension,&filename[position+1],32);
	}

	return extension;
}

/**************************************************************************************************/

int copy_file(char *old_filename, char  *new_filename)
{
		FILE  *ptr_old, *ptr_new;
		int  a;

		ptr_old = fopen(old_filename, "rb");
		ptr_new = fopen(new_filename, "wb");

		while(1)
		{
			a  =  fgetc(ptr_old);
			if(!feof(ptr_old))
				fputc(a, ptr_new);
			else
				break;
		}

		fclose(ptr_new);
		fclose(ptr_old);
		return  0;
}

int copy_file_append(char *old_filename, char  *new_filename)
{
		FILE  *ptr_old, *ptr_new;
		int  a;

		ptr_old = fopen(old_filename, "rb");
		ptr_new = fopen(new_filename, "a");

		while(1)
		{
			a  =  fgetc(ptr_old);
			if(!feof(ptr_old))
				fputc(a, ptr_new);
			else
				break;
		}

		fclose(ptr_new);
		fclose(ptr_old);
		return  0;
}

void WriteFile(char *outFileName, char *data, LONG offset, LONG fileSize)
{
	FILE *file;
	//long w1 = 0;
	
	file = fopen(outFileName, "wb");
	if (!file)
	{
		//printf("Unable to open file %s", outFileName);		
	}
	else
	{       
        fseek(file, offset, SEEK_SET);	   
	    fwrite(data , 1, fileSize, file);
	    //printf("\nDone writing file: %s, Length: %d\n\n", outFileName,  fileSize);
        }    
	fclose(file);
}

/**************************************************************************************************/

// Convert an image to a bitmap
void image_to_bitmap(struct Image *image, struct BitMap *bitmap, short depth )
{
short plane=1;
long planesize;

// Initialise bitmap
InitBitMap( bitmap, depth, image->Width, image->Height );

// Get plane size
planesize = ((image->Width+15)>>4) * image->Height;

// First plane pointer
bitmap->Planes[0] = (PLANEPTR)image->ImageData;

// More than 1 plane?
if	(image->Depth > 1)
	{
	// Get second plane pointer
	bitmap->Planes[plane++] = (PLANEPTR)(image->ImageData+planesize);

	// If it's 3 planes, under 39, we can remap it
	//if	(image->Depth==3 && GfxBase->LibNode.lib_Version>=39 && data->remap)
		//{
		//for	(plane = 2; plane < depth; plane++)
			//bitmap->Planes[plane] = (PLANEPTR)(image->ImageData + (planesize << 1));
		//}

	// More than 2 planes?
	//else
	if	(image->Depth>2)
		{
		for	(plane = 2; plane<image->Depth; plane++)
			bitmap->Planes[plane] = (PLANEPTR)(image->ImageData + planesize * plane);
		}
	}

// Clear extra planes
for	(; plane < depth; ++plane)
	bitmap->Planes[plane] = 0;
}

/**************************************************************************************************/

//void ConvertRGBAImage2ARGB(ULONG *RGBAImage, unsigned char *ARGBTemp, LONG wd, LONG ht, char *PixelFormat)
void ConvertRGBAImage2ARGB(unsigned char *RGBAImage, unsigned char *ARGBTemp, LONG wd, LONG ht, char *PixelFormat)
{
    //Convert RGBA to BGRA to Save As 32bit BMPX.
    //Convert RGBA to ARGB to Save As 32bit BMPX or 32bit PNG.
        
    int x, y;    
    int source = 0;
    unsigned char a,r,g,b;    
    
    if(strcmp(PixelFormat, "BGRA") == 0)
    {
        for (y = 0; y < ht; y++)
        { 
            for (x = 0; x < wd; x++)
            {
                //r = ((RGBAImage[source]) & 0x000000ff);
                //g = ((RGBAImage[source]) & 0x0000ff00) >> 8;
                //b = ((RGBAImage[source]) & 0x00ff0000) >> 16;
                //a = ((RGBAImage[source]) & 0xff000000) >> 24;
                
                //Convert ULONG to BYTES
                //r = ((RGBAImage[source]) & 0xff);          //TIFFGetR ((rgba) & 0xff)
                //g = (((RGBAImage[source]) >> 8) & 0xff);  //TIFFGetG (((rgba) >> 8) & 0xff)
                //b = (((RGBAImage[source]) >> 16) & 0xff); //TIFFGetB (((rgba) >> 16) & 0xff)
                //a = (((RGBAImage[source]) >> 24) & 0xff); //TIFFGetA (((rgba) >> 24) & 0xff) 

                r = (RGBAImage[source]);          
                g = (RGBAImage[source+1]);  
                b = (RGBAImage[source+2]); 
                a = (RGBAImage[source+3]);              
                                
                //Convert RGBA to BGRA
                ARGBTemp[source] = b; 
                ARGBTemp[source+1] = g; 
                ARGBTemp[source+2] = r; 
                ARGBTemp[source+3] = a; 
                
                //Convert RGBA to BGRA
                //ARGBTemp[source] = r; 
                //ARGBTemp[source+1] = g; 
                //ARGBTemp[source+2] = b; 
                //ARGBTemp[source+3] = a;
		source += 4;
            }
        }
    }
    else if(strcmp(PixelFormat, "ARGB") == 0)
    {
        for (y = 0; y < ht; y++)
        {          
            for (x = 0; x < wd; x++)
            {
                //Convert ULONG to BYTES
                //r = ((RGBAImage[source]) & 0xff);          //TIFFGetR ((rgba) & 0xff)
                //g = (((RGBAImage[source]) >> 8) & 0xff);  //TIFFGetG (((rgba) >> 8) & 0xff)
                //b = (((RGBAImage[source]) >> 16) & 0xff); //TIFFGetB (((rgba) >> 16) & 0xff)
                //a = (((RGBAImage[source]) >> 24) & 0xff); //TIFFGetA (((rgba) >> 24) & 0xff) 

                r = (RGBAImage[source]);          
                g = (RGBAImage[source+1]);  
                b = (RGBAImage[source+2]); 
                a = (RGBAImage[source+3]); 
                
                //Convert RGBA to ARGB 
                ARGBTemp[source] = a;
                ARGBTemp[source+1] = r;
                ARGBTemp[source+2] = g;
                ARGBTemp[source+3] = b;
                source += 4;
            }
        }        
    }    
}

/**************************************************************************************************/

void ConvertRGBAImage2RGB(unsigned char *RGBAImage, unsigned char *RGBTemp, LONG wd, LONG ht)
{    
	//Convert RGBA to RGB to Save As 24bit PNG.
        
	int x, y;    
	int source = 0;
	int target = 0;
	unsigned char r,g,b; 
	
	for (y = 0; y < ht; y++)
	{ 
		for (x = 0; x < wd; x++)
		{    
			r = (RGBAImage[source]);          
			g = (RGBAImage[source+1]);  
			b = (RGBAImage[source+2]); 
			//a = (RGBAImage[source+3]); 
			
			//Convert RGBA to RGB
			RGBTemp[target] = r;
			RGBTemp[target+1] = g;
			RGBTemp[target+2] = b;
			//RGBTemp[target+3] = a;			
						
			target += 3;
			source += 4;

			//int arrayOffset = (y * wd + x);
			//int rgba = RGBAImage[arrayOffset];
			//r = (unsigned char)TIFFGetR(rgba);
			//g =  (unsigned char)TIFFGetG(rgba);
			//b =  (unsigned char)TIFFGetB(rgba);
			//a =  (unsigned char)TIFFGetA(rgba);
			
			//RGBImage[source] = r; 
			//RGBImage[source+1] = g; 
			//RGBImage[source+2] = b; 
			//RGBImage[x+3] = a;
			
			//source += 3;
		}
	}
}

/**************************************************************************************************/

/*************
 * DESCRIPTION:   Loads a tiff image with tiff library.
 * INPUT:         filename    name of picture to load
 * OUTPUT:        FALSE if failed
 *************/
BOOL Load_TIFF(char *fname)
{
	//ULONG w, h;
	//ULONG* raster;
	BOOL retval = FALSE;
	//unsigned char *src, *dst;

        //Check file signature	
	
	EasyRequest(NULL, &ResultReq, NULL, "Open TIFF Image");
	
	//#include <tiffio.h>
	//#define TIFFGetR(abgr) ((abgr) & 0xff)
	//#define TIFFGetG(abgr) (((abgr) >> 8) & 0xff)
	//#define TIFFGetB(abgr) (((abgr) >> 16) & 0xff)
	//#define TIFFGetA(abgr) (((abgr) >> 24) & 0xff)
	//int TIFFReadRGBAImage(TIFF *tif, u_long width, u_long height, u_long *raster, int stopOnError)
	//int TIFFReadRGBAImageOriented(TIFF *tif, u_long width, u_long height, u_long *raster, int orientation, int stopOnError)
	
	//TIFF* tif = TIFFOpen("Ram Disk:viking_ship.tif","r");
	TIFF* tif = TIFFOpen(fname,"r");
	if (tif)
	{		
		ULONG w, h;        
        //ULONG row;
        //tdata_t buf;
        ULONG *buf;
        LONG scanline;
        char *pictureData;
	    LONG pconfig, pmetric;
        
        pictureData = malloc(200 * sizeof(char));

        TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
  	    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
        TIFFGetField(tif, TIFFTAG_PLANARCONFIG, &pconfig);
        TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &pmetric);
        

  	    sprintf(pictureData, "\nPicture Data: Width %d x Height %d\n", w, h);
        //sprintf(pictureData, "\nPicture Data: Width %d x Height %d , BitDepth: %d\n", w, h, bit);

        EasyRequest(NULL, &ResultReq, NULL, pictureData);
        free(pictureData);
	       pictureData = NULL;
           
        //int offset = 0;
        scanline = TIFFScanlineSize(tif);
	    //LONG RGBScanlineSize = (w * h * 3);
	    buf = (ULONG *)_TIFFmalloc(scanline);		
		
		
        if (pconfig == PLANARCONFIG_CONTIG) 
        {
            //for (row = 0; row < h; row++)
            //{
                //TIFFReadScanline(tif, buf, row);
                //...process scanline data...
                //unsigned char *RGBScanline = ConvertScanline(buf);
			    
			    //void WriteFile(char *outFileName, char *data, LONG offset, LONG fileSize)
			    //WriteFile("Ram Disk:TiffData.dat", RGBScanline, offset, RGBScanlineSize);
                //WriteFile("Ram Disk:TiffData.dat", (unsigned char *)buf, offset, scanline);
                //offset += RGBScanlineSize;
            //}
        }
        else if (pconfig == PLANARCONFIG_SEPARATE) 
        {
            uint16 nsamples;

            TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &nsamples);
            //for (s = 0; s < nsamples; s++)
            //for (row = 0; row < h; row++)
                //TIFFReadScanline(tif, buf, row, s);
                //...process scanline data...
        }
        _TIFFfree(buf);               
        

        uint32* raster;            
        uint32 npixels = w*h;
        
        raster=(uint32 *) _TIFFmalloc(npixels *sizeof(uint32));

        //TIFFReadRGBAImage(tif, w, h, raster, 0); //ORIENTATION_TOPLEFT
        //if (TIFFReadRGBAImageOriented(tif, width, height, raster, ORIENTATION_TOPLEFT, 1)) {
        TIFFReadRGBAImageOriented(tif, w, h, raster, ORIENTATION_TOPLEFT, 0);

        //WriteFile("Ram Disk:TiffData.dat", (unsigned char *)raster, 0, npixels * 4);

        unsigned char *RGBImage = NULL;
        //unsigned char *ARGBImage = NULL;
        RGBImage = (unsigned char *)AllocVec(npixels * 3,MEMF_ANY);
        //ARGBImage = (unsigned char *)AllocVec(npixels * 4,MEMF_ANY);

        //Convert RGBA to BGRA to Save As 32bit BMPX. //Only works with TIFFReadRGBAImage.
	    //EasyRequest(NULL, &ResultReq, NULL, "Load_TIFF(): Convert RGBA to BGRA");   
        //ConvertRGBAImage2ARGB( (unsigned char *)raster, ARGBImage, w, h, "BGRA" ); 
        //WriteFile("Ram Disk:TiffData_32.dat", ARGBImage, 0, npixels * 4);
        
	    //Convert RGBA to ARGB to Save As 32bit BMPX or 32bit PNG. 
	    //EasyRequest(NULL, &ResultReq, NULL, "Load_TIFF(): Convert RGBA to ARGB");
        //ConvertRGBAImage2ARGB( (unsigned char *)raster, ARGBImage, w, h, "ARGB" );        
	    //WriteFile("Ram Disk:TiffData_32.dat", ARGBImage, 0, npixels * 4);
	
	    //showImage(ARGBImage, w, h, 32, fname);
        
        //** Force to Save As 32bit PNG **//
        //EasyRequest(NULL, &ResultReq, NULL, "Load_TIFF(): Save ARGB Image by Datatype");
        //BOOL ret = savepix_by_datatype(ARGBImage, "Ram Disk:Image_32.png", w, h, "png");	
	
        
        //Convert RGBA to RGB to Save As 24bit PNG.
	    EasyRequest(NULL, &ResultReq, NULL, "Load_TIFF(): Convert RGBA to RGB");
	    ConvertRGBAImage2RGB( (unsigned char *)raster, RGBImage, w, h ); 
	    //void Remove_Alpha(unsigned char *pixels_buffer, unsigned char *pixels_buffer2, int width, int height)
	    //Remove_Alpha( (unsigned char *)raster, RGBImage, w, h);
	
	    showImage(RGBImage, w, h, 24, fname);
        
        //** Force to Save As 24bit PNG **//
        //EasyRequest(NULL, &ResultReq, NULL, "Load_TIFF(): Save RGB Image by Datatype");
        //ret = savepix_by_datatype(RGBImage, "Ram Disk:Image_24.png", w, h, "png");

        _TIFFfree(raster);

        //Write DTImage to file using DTWM_RAW method.
		
		//showimage(buf...); ??
		
		//What you have now is an array scanline) of pixel values, one for each pixel in the image, and each pixel is a 4-byte value.  
		//Each of the bytes represent a channel of the pixel value (RGBA).  Each channel is represented by an integer value from 0 to 255.
		
		//To get each of the individual channel of a pixel use the function:

		//char X=(char )TIFFGetX(raster[i]);  // where X can be the channels R, G, B, and A.
		// i is the index of the pixel in the raster.
		//Important: Remember that the origin of the raster is at the lower left corner.  You should be able to figure out how the image 
		//is stored in the raster given that the pixel information is stored a row at a time!

        TIFFClose(tif);
        retval = TRUE;
	}
	
	
	return retval;
}

/**************************************************************************************************/

/*************
 * DESCRIPTION:   Saves a tiff image with tiff library.
 * INPUT:                  filename    name of picture to load
 * OUTPUT:              FALSE if failed
 *************/
//BOOL Save_TIFF(char *output_file, unsigned char *image, struct BitMapHeader *bmhd)
BOOL Save_TIFF(char *output_file, unsigned char *image, int wd, int ht, int dp)
{
	//BOOL err;
    ULONG row;
    //ULONG wd, ht, dp;
	UWORD photometric;
	UWORD compression;
	UWORD samplesperpixel;
	//BOOL hascolormap = FALSE;
	//UWORD = uint16
	//IPTR sourcetype;
	//BPTR FileHandle;	
	BOOL retval = FALSE;	
    //STRPTR namebuffer = NULL;
	//struct BitMapHeader *bmhd;
	ULONG rowsperstrip = (ULONG) -1;
	
	//EasyRequest(NULL, &ResultReq, NULL, "Save_TIFF(): Beginning. (24 bit mode)");
	
	//Creating a new TIFF image file;
	//You must open the file with the following line to write the tiff file, 
	//this will create the file if it does not exist, and replace the file if it does.

	EasyRequest(NULL, &ResultReq, NULL, "Save_TIFF(): Open TIFF for Writing...");
	
	// Open Tiff file to get new filehandle  for writing.
	TIFF* tif = TIFFOpen(output_file, "w");
	if (tif)
	{
		// Get width, height, depth for the output file.
		//wd = bmhd->bmh_Width;
		//ht = bmhd->bmh_Height;
		//dp = bmhd->bmh_Depth;
		
		// Get compression sheme for the output file.
		compression = COMPRESSION_LZW;
		//compression = COMPRESSION_PACKBITS;
		//compression = COMPRESSION_NONE;	
		
		
		if (dp == 24)
		{
			samplesperpixel = 3; // there is no alpha channel
			photometric = PHOTOMETRIC_RGB;			
		}
		else
		{
			retval = FALSE;	
		}
		
		/* ----------------------------------------------------- */
		/*  Prepare output file.                                               */
		/* ----------------------------------------------------- */
		
		EasyRequest(NULL, &ResultReq, NULL, "Save_TIFF(): Prepare Output File...");
		
		TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, wd);
		TIFFSetField(tif, TIFFTAG_IMAGELENGTH, ht);
		TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
		TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, samplesperpixel);
		TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
		TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
		TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, photometric);
		//TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, 1);	
		TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize(tif, rowsperstrip));
		TIFFSetField(tif, TIFFTAG_COMPRESSION, compression);
		
		/* ----------------------------------------------------- */
		/*  Prepare image data.                                              */
		/* ----------------------------------------------------- */
		
		EasyRequest(NULL, &ResultReq, NULL, "Save_TIFF(): Prepare Image Data...");
		
		//int x;
        int y;
		int offset;
		int rgbSize;
		unsigned char *rowbuffer;
		rowbuffer = (unsigned char *) _TIFFmalloc(wd * samplesperpixel);	
		if(!rowbuffer)
		{
			EasyRequest(NULL, &ResultReq, NULL, "Save_TIFF(): Failed to Allocate Row Buffer.");
			retval = FALSE;
		}

		// Write out the image data line by line.
		EasyRequest(NULL, &ResultReq, NULL, "Save_TIFF(): Write TIFF Scanlines...");
		for (y = 0; y < ht; y++ )
		{
			row = y;
			rgbSize = (wd* 3);
			offset = (y * wd* 3);			
			
			//Copy RGB Rows from RGB Image Buffer to Row Buffer						
			memcpy(rowbuffer, image + offset, rgbSize);				
			TIFFWriteScanline(tif, rowbuffer, row, 0);
			
		}
		_TIFFfree(rowbuffer);
		TIFFClose(tif);
		retval = TRUE;
		//EasyRequest(NULL, &ResultReq, NULL, "\nSuccessfully Saved 24bit Picture File.\n");
	}
	else
	{
		EasyRequest(NULL, &ResultReq, NULL, "Save_TIFF(): Failed to Open File for Writing.");
		retval = FALSE;
		return retval;
	}
	//EasyRequest(NULL, &ResultReq, NULL, "Save_TIFF(): Ending. (24 bit mode)");
	
	
	return retval;
}

/**************************************************************************************************/

/*************
 * DESCRIPTION:   Loads an image using the datatypes
 * INPUT:         filename   name of picture to load
 * OUTPUT:        FALSE if failed
 *************/
BOOL LoadAsDatatype(char *fname)
{    
    int width, height, depth;
    RGBImage *in_pic = NULL;
    struct BitMapHeader *bmhd; 
    BOOL retval = FALSE;    
     
    retval = testImage_by_datatype(fname);
	if(retval == TRUE)
	{  
    //Check if File Exists
    //BOOL res = EXISTS(fname);
    //if(res == TRUE)
    //{
       //Find Width, Height & BitDepth of Image.
	   bmhd = readbmhd_by_datatype(fname);
	   width = bmhd->bmh_Width;
	   height = bmhd->bmh_Height;
	   depth = bmhd->bmh_Depth;
    //}
	
	if((width > 0) && (height > 0))
	{
	   //Show BitDepth & Image Data	
        char *pictureMessage;
        pictureMessage = malloc(200 * sizeof(char));

	   //Get FileName and Extension from Path.
        char* picTitle = getFileNameFromPath(fname);

        //Detect Image Type from File Signature    
        char *picType = DetectImageType_by_datatype(fname);
    
        sprintf(pictureMessage, "\nPicture Type: %s\n\nPicture Name: %s\n\nPicture Data: Width %d x Height %d , BitDepth: %d\n", picType, picTitle, bmhd->bmh_Width, bmhd->bmh_Height, bmhd->bmh_Depth);
	    //EasyRequest(NULL, &ResultReq, NULL, pictureMessage);
	    //Free Resources for pictureMessage
	    free(pictureMessage);
	    pictureMessage = NULL;

	    in_pic = (RGBImage *)AllocVec(sizeof(RGBImage),MEMF_ANY);
	    
	    //For Displaying ILBM Bitmap Images.
        if(depth < 8)
            depth = 8;
            
	    
	    if(depth == 8)
        {
	        retval = readpic_by_datatype_alpha(in_pic, fname);
	        if(retval == TRUE)
	        {
	            showImage(in_pic->Data, width, height, depth, fname);

                return TRUE;
            }
        }
        else if(depth == 24)	    
	    {
	        retval = readpic_by_datatype(in_pic, fname);
	        if(retval == TRUE)
	        {
	            showImage(in_pic->Data, width, height, depth, fname);
	        
                return TRUE;
            }	        
        }
        else if(depth == 32)
        {
	        retval = readpic_by_datatype_alpha(in_pic, fname);
            
            //if(picType == "bmp")
			if (strcmp(picType, "bmp") == 0)
	        {
                BPTR fileHandle = Open(fname, MODE_OLDFILE); //MODE_OLDFILE
	            Seek(fileHandle,54,OFFSET_BEGINNING);
	            Read(fileHandle,in_pic->Data,(in_pic->Width*4*in_pic->Height)); //BGRA
	            UBYTE tempRGB;
	            int imageIdx = 0;
                for (imageIdx = 0;imageIdx < (in_pic->Width*4*in_pic->Height);imageIdx+=4)
                {
                    tempRGB = in_pic->Data[imageIdx]; //blue
				    in_pic->Data[imageIdx] = in_pic->Data[imageIdx + 3]; //alpha
				    in_pic->Data[imageIdx + 3] = tempRGB; //blue

				    tempRGB = in_pic->Data[imageIdx+1]; //green
				    in_pic->Data[imageIdx+1] = in_pic->Data[imageIdx + 2]; //red
				    in_pic->Data[imageIdx + 2] = tempRGB; //green
                }
                Flip_Vertical(in_pic->Data, in_pic->Width, in_pic->Height, 4);
            }
	        if(retval == TRUE)
	        {
	            showImage(in_pic->Data, width, height, depth, fname);

                return TRUE;
            }
        }
        else
        {
            EasyRequest(NULL, &ResultReq, NULL, "Error! Wrong BitDepth.");
        }	
    } 
    else
    {
       EasyRequest(NULL, &ResultReq, NULL, "Can't Open Picture File!"); 
    }
    
    } //if Test Image 
    else
    {
       EasyRequest(NULL, &ResultReq, NULL, "Can't Open Picture File!");
    }
    
        if (in_pic)
		{
			if (in_pic->Data) FreeVec(in_pic->Data);
			FreeVec(in_pic);
		}
    
    
    return FALSE;
}

/**************************************************************************************************/

/*************
 * DESCRIPTION:   Saves an image using picture datatypes. 
 * INPUT:         old_fname    name of picture to load
 * INPUT:         filename     name of picture to save
 * INPUT:         pType        type of picture to save
 * OUTPUT:        FALSE if failed
 *************/

//OpenDTO, ConvertDTO, SaveDTO, CloseDTO
BOOL SaveAsDatatype(char *old_fname, char *filename, char *pType)
{
        //Save the image to file by BitDepth.
        struct BitMapHeader *bmhd;
        //UBYTE *Cmap = NULL;
        UBYTE rgb[256 * 3];
        //struct ColorRegister *colorMap;
        //BITMAPINFOHEADER bitmapInfo;
        //Object *DTImage = NULL;
        BitMapImage *BMI = NULL;
        RGBImage *in_pic = NULL; 
        //*out_pic = NULL;

        in_pic = (RGBImage *)AllocVec(sizeof(RGBImage),MEMF_ANY);
	    //out_pic = (RGBImage *)AllocVec(sizeof(RGBImage),MEMF_ANY);
	    //BMI = (BitMapImage *)AllocVec(sizeof(BitMapImage),MEMF_ANY);
        
	    //Find Width, Height & BitDepth of Image.
	    bmhd = readbmhd_by_datatype(old_fname);
	    //char Type[4];
	    //int width = bmhd->bmh_Width;
	    //int height = bmhd->bmh_Height;
	    int bit = bmhd->bmh_Depth;
	    
        //For ILBM Images.
        if(bit < 8)
            bit = 8;
	    
	    
	    //Read Image from File by BitCount.        
	    if(bit == 24)	    
	    {
            //EasyRequest(NULL, &ResultReq, NULL, "readpic_by_datatype");
	        readpic_by_datatype(in_pic, old_fname); //RGB Format
        }        
        else if(bit == 32)        
        {
            //EasyRequest(NULL, &ResultReq, NULL, "readpic_by_datatype_alpha");
	        readpic_by_datatype_alpha(in_pic, old_fname); //RGB Format
        }
        else if(bit == 8)
        {
           int numcolors = 256;
           //EasyRequest(NULL, &ResultReq, NULL, "readpic_by_datatype_lut8");
	       readpic_by_datatype_lut8(BMI, old_fname);

           //Convert Cmap to Byte Array.

           int i;
           struct ColorRegister *cols = Cmap_by_datatype(old_fname);
           for( i = 0; i < numcolors ; i++ )
           {
               rgb[i * 3] = cols[i].red;
               rgb[(i * 3)+1] = cols[i].green;
               rgb[(i * 3)+2] = cols[i].blue;
           }           
        }
        else
        {            
            EasyRequest(NULL, &ResultReq, NULL, "\nCan't Open Picture File!\n");
        }


        //Save Image to File by BitDepth.
        
        BOOL res = FALSE;
        if(bit == 24)        
	    {
            //EasyRequest(NULL, &ResultReq, NULL, "savepic_by_datatype");
            res = savepic_by_datatype(in_pic, filename, pType); //RGB
        }
        else if(bit == 32)
	    {
            //EasyRequest(NULL, &ResultReq, NULL, "savepic_by_datatype_alpha");
            res = savepic_by_datatype_alpha(in_pic, filename, pType); //ARGB
        }
        else if(bit == 8)
	    {
            //EasyRequest(NULL, &ResultReq, NULL, "savepic_by_datatype_lut8");
            res = savepic_by_datatype_lut8(BMI, rgb, filename, pType); //LUT8
        }
        else
        {            
            EasyRequest(NULL, &ResultReq, NULL, "\nCan't Save Picture File!\n");
        }

        //Check results.
        if (res == TRUE)
		{
				 EasyRequest(NULL, &ResultReq, NULL, "Successfully Saved Picture File.");
		}
		else
		{
				EasyRequest(NULL, &ResultReq, NULL, "Error! Failed to Save Picture File!");
		}


        //FreeVec causes Memory Header Not Found Error!
        if (in_pic)
		{
			if (in_pic->Data) FreeVec(in_pic->Data);
			FreeVec(in_pic);
		}
		//if (out_pic)
		//{
			//if (out_pic->Data) FreeVec(out_pic->Data);
			//FreeVec(out_pic);
		//}
		//if (BMI)
		//{
			//if (BMI->Data) FreeVec(BMI->Data);
			//if (BMI->Cmap) FreeVec(BMI->Cmap);
			//if (BMI->Cregs) FreeVec(BMI->Cregs);
			//FreeVec(BMI);
		//}
        return FALSE;
}

/**************************************************************************************************/

int showImage(char *bitmapData, int width, int height, int depth, char *fpath)
{
            //struct Window *win;
            BOOL quitme = FALSE;
    
            //Get FileName and Extension from Path.
            char* fname = getFileNameFromPath(fpath);
            char *title = fname;
            
            //Menu Items
            struct Window *win;
            struct Screen *scr;
            //struct IntuiMessage *imsg;            
            APTR vi = NULL;            
            struct Menu *menu = NULL;
            struct MenuItem *item;            
            UWORD menuNumber;
            UWORD menuNum;
            UWORD itemNum;
            //UWORD subNum;
    
            //Set minimum window size based on image size
            int winw = width;
            int winh = height;
            int winl = 2;
            int wint = 2;
			if(width < 100)
			{                
                winw = 110;                
                winh = 100;                

                winl = (110 - width) / 2;
                wint = (100 - height) / 2;
            }
            
            //Open New Window
            win = OpenWindowTags(0, WA_Title, (IPTR)title,
    	    WA_InnerWidth, winw + 4,
			WA_InnerHeight, winh + 4,
			WA_SmartRefresh, TRUE,
			WA_CloseGadget, TRUE,
			WA_DragBar, TRUE,
			WA_DepthGadget, TRUE,
			WA_Activate, TRUE,			
			WA_IDCMP, IDCMP_CLOSEWINDOW | IDCMP_VANILLAKEY | IDCMP_MENUPICK | IDCMP_MOUSEBUTTONS | IDCMP_GADGETUP |IDCMP_GADGETDOWN,
			TAG_DONE);
            
            /* Attach menus to window */            

            scr = win->WScreen;            
            if (vi = GetVisualInfoA (scr,NULL))
                if (menu = CreateMenusA (newmenu,NULL))
			     if (LayoutMenus (menu,vi,GTMN_NewLookMenus,TRUE,TAG_END))
				        SetMenuStrip (win,menu);
			
			//Must know width, height, depth of bitmap data.
			if (win)
            {
                quitme = FALSE;
                if (GetBitMapAttr(win->WScreen->RastPort.BitMap, BMA_DEPTH) >= 15)
	            {
                    if(depth == 8)
                    {
                        //Display as 32bit ARGB
                        WORD dp = 32;
                        WritePixelArrayAlpha(bitmapData, 0, 0, width * dp / 8, win->RPort, win->BorderLeft + winl, win->BorderTop + wint, width, height, RECTFMT_ARGB);
                    }
                    else if(depth == 24)                    
                    {                        
                        WritePixelArray(bitmapData, 0, 0, width * depth / 8, win->RPort, win->BorderLeft + winl, win->BorderTop + wint, width, height, RECTFMT_RGB);
                    }
                    else if(depth == 32)
                    {
                        WritePixelArrayAlpha(bitmapData, 0, 0, width * depth / 8, win->RPort, win->BorderLeft + winl, win->BorderTop + wint, width, height, RECTFMT_ARGB);
                    }
                    else
                    {                        
                        EasyRequest(NULL, &ResultReq, NULL, "Error! Unsupported PixelFormat.");
                    }
                }
                else
                {
                    EasyRequest(NULL, &ResultReq, NULL, "Error! Wrong Screen Depth.");
                }
            //} //if win
            
                while(!quitme)
                {
	    	          struct IntuiMessage *msg;                      
	    	          WaitPort(win->UserPort);
		              while((msg = (struct IntuiMessage *)GetMsg(win->UserPort)))
		              {
		                  switch(msg->Class)
		                  {
		    	            case IDCMP_CLOSEWINDOW:
			                quitme = TRUE;
			                break;

			                case IDCMP_VANILLAKEY:
			                if (msg->Code == 27) quitme = TRUE;
			                break;
                            
                            case IDCMP_MENUPICK:
                            menuNumber = msg->Code;
				            for (item = ItemAddress(menu,msg->Code); item; item = ItemAddress(menu,item->NextSelect))
					        {
					            //Printf ("menu selected: %s\n",((struct IntuiText *)item->ItemFill)->IText);
                                /* process the item here! */
                                menuNum = MENUNUM(menuNumber);
                                itemNum = ITEMNUM(menuNumber);
                                //subNum  = SUBNUM(menuNumber);

                                char *itemName;
                                itemName = malloc(60 * sizeof(char));
	                            sprintf(itemName, "Item Name: %s", ((struct IntuiText *)item->ItemFill)->IText);
					            //EasyRequest(NULL, &ResultReq, NULL, itemName);
					            free(itemName);
					            itemName = NULL;

					            char *itemNumber;
					            itemNumber = malloc(60 * sizeof(char));
					            sprintf(itemNumber, "Item Num: %d", itemNum);
					            //EasyRequest(NULL, &ResultReq, NULL, itemNumber);
					            free(itemNumber);
					            itemNumber = NULL;
                                
                                //MSG_MEN_PROJECT_FILE:
                                if ((menuNum == 0) && (itemNum == 0))
                                {
                                    //Show Full Path of File.
                                    EasyRequest(NULL, &ResultReq, NULL, fpath);
                                }
                                //MSG_MEN_PROJECT_OPEN
                                //if ((menuNum == 0) && (itemNum == 2))
                                //MSG_MEN_PROJECT_SAVEAS
                                if ((menuNum == 0) && (itemNum == 2))
                                {
                                    //EasyRequest(NULL, &ResultReq, NULL, "Save As"); 
                                    //Get filename to save.
                                    char *filePart;
                                    STRPTR fileName = GetNewFileName(&filePart, "Save File");

                                    if((!fileName) || (strlen(filePart) == 0))
                                    {
                                        EasyRequest(NULL, &ResultReq, NULL, "\nNo File Selected!\n");
                                    }
                                    else
                                    {
                                        //Check If fileName EXISTS.
                                        BOOL res = EXISTS(fileName);
                                        if(res == TRUE)
                                        {
                                            //Prompt to overwrite existing file. Only ask one time!
                                            char *question = "\nFile Exists. Would you like to overwrite?\n";
                                            if (EasyRequest(NULL, &YesNoReq, NULL, question) == YES)
                                            {
                                                //Do Nothing. File with be Replaced.
                                            }
                                            else
                                            {
                                                //Get a New File Name.
                                                fileName = GetNewFileName(&filePart, "Save File");
                                            }
                                        }
                                        //Use copy_file
                                        copy_file(fpath, fileName);                                        
                                    }
                                }
                                //MSG_MEN_PROJECT_SAVEAS_TIFF
                                if ((menuNum == 0) && (itemNum == 3))
                                {
					                //Get filename to save.
                                    char *filePart;
                                    STRPTR fileName = GetNewFileName(&filePart, "Save File");

                                    if((!fileName) || (strlen(filePart) == 0))
                                    {
                                        EasyRequest(NULL, &ResultReq, NULL, "\nNo File Selected!\n");
                                    }
                                    else
                                    {                                        
                                        //Check If fileName EXISTS.
                                        BOOL res = EXISTS(fileName); 
                                        if(res == TRUE)
                                        {                                            
                                            //Prompt to overwrite existing file. Only ask one time!
                                            char *question = "\nFile Exists. Would you like to overwrite?\n";
                                            if (EasyRequest(NULL, &YesNoReq, NULL, question) == YES)
                                            {
                                                //Do Nothing. File with be Replaced.                                                
                                            }
                                            else
                                            {
                                                //Get a New File Name.
                                                fileName = GetNewFileName(&filePart, "Save File");
                                            }
                                        }                                        
                                        
					                //BOOL ret = SaveAsDatatype(fpath, fileName, "tiff");
							        //BOOL SaveAsDatatype(char *old_fname, char *filename, char *pType)
					                BOOL ret = Save_TIFF(fileName, bitmapData, width, height, 24);
					                //BOOL ret = savepix_by_datatype(bitmapData, fileName, width, height, "tiff");
					                if (ret == TRUE)		
						                EasyRequest(NULL, &ResultReq, NULL, "Successfully Saved Picture File.");		
					                else
						                EasyRequest(NULL, &ResultReq, NULL, "Error! Failed to Save Picture File!");                                                                                   
                                    }
				                }
                                //MSG_MEN_PROJECT_SAVEAS_PNG
                                if ((menuNum == 0) && (itemNum == 4))
                                {
                                    //EasyRequest(NULL, &ResultReq, NULL, "Save As PNG");
                                    //Get filename to save.
                                    char *filePart;
                                    STRPTR fileName = GetNewFileName(&filePart, "Save File");

                                    if((!fileName) || (strlen(filePart) == 0))
                                    {
                                        EasyRequest(NULL, &ResultReq, NULL, "\nNo File Selected!\n");
                                    }
                                    else
                                    {                                        
                                        //Check If fileName EXISTS.
                                        BOOL res = EXISTS(fileName);
                                        BOOL append = FALSE;
                                        //Get file extension.
                                        char *ext = get_ext(fileName);
                                        if(res == TRUE)
                                        {                                            
                                            //Prompt to overwrite existing file. Only ask one time!
                                            char *question = "\nFile Exists. Would you like to overwrite?\n";
                                            if (EasyRequest(NULL, &YesNoReq, NULL, question) == YES)
                                            {
                                                //Do Nothing. File with be Replaced.
                                                //Check if fileName EXISTS & Ext == ".info"
                                                if(strnicmp((const char *)ext, "info", 4)==0)
                                                {
                                                    char *question = "\nINFO File Exists. Would you like to Append?\n";
                                                    if (EasyRequest(NULL, &YesNoReq, NULL, question) == YES)
                                                    {
                                                        //if YES=Overwrite & Append .info, append = TRUE;
                                                        append = TRUE;
                                                    }
                                                }
                                            }
                                            else
                                            {
                                                //Get a New File Name.
                                                fileName = GetNewFileName(&filePart, "Save File");
                                            }
                                        }
                                        
                                        //Use picture datatypes
                                        if( append == TRUE )
                                        {                                            
                                             SaveAsDatatype(fpath, "Ram Disk:Im1.dat", "png");
                                             copy_file_append("Ram Disk:Im1.dat", fileName);
                                        }
                                        else
                                        {
                                            SaveAsDatatype(fpath, fileName, "png");
                                        }                                            
                                    }
                                }
                                //MSG_MEN_PROJECT_ABOUT
                                if ((menuNum == 0) && (itemNum == 6))
                                {
                                    /* Display About Message Window */                                    

                                    //Get the picture information.
                                    char *pictureMessage;
                                    pictureMessage = malloc(260 * sizeof(char));
                                                                        
                                    //char *pictureType = DetectImageType(fpath);
                                    char *pictureType = DetectImageType_by_datatype(fpath);                                    

                                    //if (pictureType == "iff")
									if (strcmp(pictureType, "iff") == 0)
                                    {
                                        pictureType = "ILBM";
                                    }
                                    
                                    //Find Width, Height & BitDepth of Image.
                                    struct BitMapHeader *bmhd;
                                    bmhd = readbmhd_by_datatype(fpath);
	                                int wd = bmhd->bmh_Width;
	                                int ht = bmhd->bmh_Height;
	                                int bit = bmhd->bmh_Depth;
                                    
                                    //Show version string with date.                                    
                                    char *codeVersion = ("Icaros Viewer v1.0.2 - Code Base: 10-01-19.");
	                                //Show Width x Height, BitDepth of Current Image	                                
                                    sprintf(pictureMessage, "\nPicture Type: %s\n\n%s\n\nPicture Data: Width %d x Height %d , BitDepth: %d\n", pictureType, codeVersion, wd, ht, bit);

                                    EasyRequest(NULL, &ResultReq, NULL, pictureMessage);
                                    free(pictureMessage);
	                                pictureMessage = NULL;                                    
                                }
                                //MSG_MEN_PROJECT_HELP
                                //if ((menuNum == 0) && (itemNum == 7))
                                //{
                                    //Instructions for Program Usage
                                    //EasyRequest(NULL, &ResultReq, NULL, "Show Instructions");
                                //}
                                //MSG_MEN_PROJECT_QUIT
                                if ((menuNum == 0) && (itemNum == 8))
                                {
                                    quitme = TRUE;
                                }
                            }
				            break;
			             } //Switch
		                 ReplyMsg((struct Message *)msg);
		              } //While (Msg)
                } //while(!quitme)
                
                //Remove Menus & Free Resources
                ClearMenuStrip (win);                
	            if (menu) FreeMenus (menu);
	            if (vi)   FreeVisualInfo (vi);                
                CloseWindow (win);
        } //if win
        
        return 0;
}

/**************************************************************************************************/

/**************************************************************************************************/

struct ColorRegister *Cmap_by_datatype(char *file_name)
{    
    struct ColorRegister *colormap;
	Object *DTImage = NULL;	

	DTImage = NewDTObject(	file_name,
							(DTA_SourceType),	DTST_FILE,
							(DTA_GroupID),		GID_PICTURE,
							(PDTA_Remap),		FALSE,
							(OBP_Precision),	PRECISION_EXACT,
							TAG_DONE);

	if (DTImage)
	{
        /* Get color palette */
        GetDTAttrs(DTImage, PDTA_ColorRegisters, &colormap, TAG_DONE );       
    }
    DisposeDTObject( DTImage );

    return colormap;
}

/*********************************************************************************************/

int CMAP2Object(Object *o, UBYTE *rgb, ULONG rgbsize )
{
    struct ColorRegister *acm;
    ULONG                *acregs;
    IPTR                 nc;

    //D(bug("[anim.datatype] %s()\n", __PRETTY_FUNCTION__));

    /* file has this many colors (e.g. each color has one byte per R,B,G) */
    nc = rgbsize / 3UL;

    SetDTAttrs( o, NULL, NULL, PDTA_NumColors, nc, TAG_DONE );

    /* Get color context */
    if( GetDTAttrs( o,
                    PDTA_ColorRegisters, (&acm),
                    PDTA_CRegs,          (&acregs),
                    PDTA_NumColors,      (&nc),
                    TAG_DONE ) == 3 )
    {
      /* All valid ? */
      if( acm && acregs && nc )
      {
        ULONG i;

        for( i = 0UL ; i < nc ; i++, acm++ )
        {
          acm -> red   =  *rgb++;
          acm -> green =  *rgb++;
          acm -> blue  =  *rgb++;

          /* Replicate the color information.
           * This surrounds an OS bug which uses the low-oreder bytes of the 32-bit colors
           * instead of the high order ones
           */
          acregs[ ((i * 3) + 0) ] = AROS_BE2LONG(((ULONG)(acm -> red))   * 0x01010101UL);
          acregs[ ((i * 3) + 1) ] = AROS_BE2LONG(((ULONG)(acm -> green)) * 0x01010101UL);
          acregs[ ((i * 3) + 2) ] = AROS_BE2LONG(((ULONG)(acm -> blue))  * 0x01010101UL);
        }        
      }
    }
    return  0;    
}

/**************************************************************************************************/

BOOL readpic_by_datatype(RGBImage *pic, char *file_name)
{
	//** Adapted from sample Code from dt2thumbs **//
    Object *DTImage = NULL;
	struct BitMapHeader *bmh;
	struct pdtBlitPixelArray bpa;

	DTImage = NewDTObject(	file_name,
							(DTA_SourceType),	DTST_FILE,
							(DTA_GroupID),		GID_PICTURE,
							(PDTA_Remap),		FALSE,
							(OBP_Precision),	PRECISION_EXACT,
							TAG_DONE);

	if (DTImage)
	{

		if (GetDTAttrs( DTImage,
					PDTA_BitMapHeader, (IPTR)&bmh,
					TAG_END ) == 1)

		{

			/* Picture struct and buffer mem allocation */
			pic->Data = (UBYTE *)AllocVec(bmh->bmh_Width * bmh->bmh_Height * 3, MEMF_ANY);
			if (pic->Data)
			{
				bpa.MethodID = PDTM_READPIXELARRAY;
				bpa.pbpa_PixelData = (APTR)pic->Data;
				bpa.pbpa_PixelFormat = PBPAFMT_RGB;
				bpa.pbpa_PixelArrayMod = bmh->bmh_Width * 3;
				bpa.pbpa_Left = 0;
				bpa.pbpa_Top = 0;
				bpa.pbpa_Width = bmh->bmh_Width;
				bpa.pbpa_Height = bmh->bmh_Height;

				DoMethodA( DTImage, (Msg)&bpa );

				pic->Width  = bmh->bmh_Width;
				pic->Height = bmh->bmh_Height;

				DisposeDTObject( DTImage );

				return TRUE;
			}
		}
		DisposeDTObject( DTImage );
	}
	return FALSE;
}

/**************************************************************************************************/

BOOL readpic_by_datatype_alpha(RGBImage *pic, char *file_name)
{
	//** Adapted from sample Code from dt2thumbs **//
    Object *DTImage = NULL;
	struct BitMapHeader *bmh;
	struct pdtBlitPixelArray bpa;

	DTImage = NewDTObject(	file_name,
							(DTA_SourceType),	DTST_FILE,
							(DTA_GroupID),		GID_PICTURE,
							(PDTA_Remap),		FALSE,
							(OBP_Precision),	PRECISION_EXACT,
							TAG_DONE);

	if (DTImage)
	{

		if (GetDTAttrs( DTImage,
					PDTA_BitMapHeader, (IPTR)&bmh,
					TAG_END ) == 1)

		{

			/* Picture struct and buffer mem allocation */
			pic->Data = (UBYTE *)AllocVec(bmh->bmh_Width * bmh->bmh_Height * 4, MEMF_ANY);
			if (pic->Data)
			{
				bpa.MethodID = PDTM_READPIXELARRAY;
				bpa.pbpa_PixelData = (APTR)pic->Data;
				bpa.pbpa_PixelFormat = PBPAFMT_ARGB;
				bpa.pbpa_PixelArrayMod = bmh->bmh_Width * 4;
				bpa.pbpa_Left = 0;
				bpa.pbpa_Top = 0;
				bpa.pbpa_Width = bmh->bmh_Width;
				bpa.pbpa_Height = bmh->bmh_Height;

				DoMethodA( DTImage, (Msg)&bpa );

				pic->Width  = bmh->bmh_Width;
				pic->Height = bmh->bmh_Height;

				DisposeDTObject( DTImage );

				return TRUE;
			}
		}
		DisposeDTObject( DTImage );
	}
	return FALSE;
}

/**************************************************************************************************/

BOOL readpic_by_datatype_lut8(BitMapImage *bmi, char *file_name)
{
    //** Adapted from sample Code from dt2thumbs **//
	Object *DTImage = NULL;
	struct BitMapHeader *bmh;
	struct pdtBlitPixelArray bpa;

	DTImage = NewDTObject(	file_name,
							(DTA_SourceType),	DTST_FILE,
							(DTA_GroupID),		GID_PICTURE,
							(PDTA_Remap),		FALSE,
							(OBP_Precision),	PRECISION_EXACT,
							TAG_DONE);

	if (DTImage)
	{		
		if (GetDTAttrs( DTImage,
                    PDTA_BitMapHeader, (IPTR)&bmh,
                    PDTA_ColorRegisters, (IPTR)&bmi->Cmap,
                    PDTA_CRegs, (IPTR)&bmi->Cregs,
					TAG_END ) == 3)
		{

			/* Picture struct and buffer mem allocation */
			bmi->Data = (UBYTE *)AllocVec(bmh->bmh_Width * bmh->bmh_Height, MEMF_ANY);
			if (bmi->Data)
			{
				bpa.MethodID = PDTM_READPIXELARRAY;
				bpa.pbpa_PixelData = (APTR)bmi->Data;
				bpa.pbpa_PixelFormat = PBPAFMT_LUT8;
				bpa.pbpa_PixelArrayMod = bmh->bmh_Width;
				bpa.pbpa_Left = 0;
				bpa.pbpa_Top = 0;
				bpa.pbpa_Width = bmh->bmh_Width;
				bpa.pbpa_Height = bmh->bmh_Height;

				DoMethodA( DTImage, (Msg)&bpa );

				bmi->Width  = bmh->bmh_Width;
				bmi->Height = bmh->bmh_Height;
				bmi->Depth = bmh->bmh_Depth;


				DisposeDTObject( DTImage );

				return TRUE;
			}
		}
		DisposeDTObject( DTImage );
	}
	return FALSE;
}

/**************************************************************************************************/

BOOL savepic_by_datatype(RGBImage *pic, char *file_name, char *type)
{
	Object *DTImage = NULL;
	struct BitMapHeader *bmhd;
	struct dtWrite dtw;
	struct pdtBlitPixelArray dtb;
	FILE *file = NULL;
	BOOL retval = FALSE;
	
	//if (type == "jpg")
	if (strcmp(type, "jpg") == 0)
	   type = "jpeg";

	
	DTImage = NewDTObject(	(APTR)NULL,
							DTA_SourceType, DTST_RAM,
							DTA_BaseName, (IPTR)type,
							PDTA_DestMode, PMODE_V43,
							TAG_DONE);
	if (!DTImage) return(FALSE);


	if (GetDTAttrs(DTImage,PDTA_BitMapHeader,(IPTR)&bmhd,TAG_DONE))
	{
		dtb.MethodID = PDTM_WRITEPIXELARRAY;
		dtb.pbpa_PixelData = pic->Data;
		dtb.pbpa_PixelFormat = PBPAFMT_RGB;
		dtb.pbpa_PixelArrayMod = pic->Width*3;
		dtb.pbpa_Left = 0;
		dtb.pbpa_Top = 0;
		dtb.pbpa_Width = pic->Width;
		dtb.pbpa_Height = pic->Height;

		bmhd->bmh_Width = pic->Width;
		bmhd->bmh_Height = pic->Height;
		bmhd->bmh_Depth = 24;
		bmhd->bmh_PageWidth = 320;
		bmhd->bmh_PageHeight = 240;

		DoMethodA(DTImage, (Msg) &dtb);

		//write datatype object to file
		if (file = Open (file_name,MODE_NEWFILE))
		{
			dtw.MethodID = DTM_WRITE;
			dtw.dtw_GInfo = NULL;
			dtw.dtw_FileHandle = file;
			dtw.dtw_Mode = DTWM_RAW;
			dtw.dtw_AttrList = NULL;

			if (DoMethodA(DTImage, (Msg) &dtw)) retval = TRUE;
		}
	}

	if (file) Close (file);
	if (DTImage) DisposeDTObject(DTImage);
	return retval;
}

/**************************************************************************************************/

BOOL savepic_by_datatype_alpha(RGBImage *pic, char *file_name, char *type)
{
	Object *DTImage = NULL;
	struct BitMapHeader *bmhd;
	struct dtWrite dtw;
	struct pdtBlitPixelArray dtb;
	FILE *file = NULL;
	BOOL retval = FALSE;
	
	//Get File Extension from File Path.
	char *ext = get_ext(file_name);

	
    //if (ext == "png")
	if (strcmp(ext, "png") == 0)
	    type = "png";
    //if (ext == "jpg")
	if (strcmp(ext, "jpg") == 0)
        //JPEG is only 24bit.
	    type = "jpeg";

	
	DTImage = NewDTObject(	(APTR)NULL,
							DTA_SourceType, DTST_RAM,
							DTA_BaseName, (IPTR)type,
							PDTA_DestMode, PMODE_V43,
							TAG_DONE);
	if (!DTImage) return(FALSE);


	if (GetDTAttrs(DTImage,PDTA_BitMapHeader,(IPTR)&bmhd,TAG_DONE))
	{
		dtb.MethodID = PDTM_WRITEPIXELARRAY;
		dtb.pbpa_PixelData = pic->Data;
		dtb.pbpa_PixelFormat = PBPAFMT_ARGB;
		dtb.pbpa_PixelArrayMod = pic->Width*4;
		dtb.pbpa_Left = 0;
		dtb.pbpa_Top = 0;
		dtb.pbpa_Width = pic->Width;
		dtb.pbpa_Height = pic->Height;

		bmhd->bmh_Width = pic->Width;
		bmhd->bmh_Height = pic->Height;
		bmhd->bmh_Depth = 32;
		bmhd->bmh_PageWidth = 320;
		bmhd->bmh_PageHeight = 240;

		DoMethodA(DTImage, (Msg) &dtb);

		//write datatype object to file
		if (file = Open (file_name,MODE_NEWFILE))
		{
			dtw.MethodID = DTM_WRITE;
			dtw.dtw_GInfo = NULL;
			dtw.dtw_FileHandle = file;
			dtw.dtw_Mode = DTWM_RAW;
			dtw.dtw_AttrList = NULL;

			if (DoMethodA(DTImage, (Msg) &dtw)) retval = TRUE;
		}
	}

	if (file) Close (file);
	if (DTImage) DisposeDTObject(DTImage);
	return retval;
}

/**************************************************************************************************/

BOOL savepic_by_datatype_lut8(BitMapImage *bmi, UBYTE *colorBytes, char *file_name, char *type)
{
	Object *DTImage = NULL;
	struct BitMapHeader *bmhd;
	struct dtWrite dtw;
	struct pdtBlitPixelArray dtb;
	//ULONG *cmap;
	FILE *file = NULL;
	BOOL retval = FALSE;
	
	//Get File Extension from File Path.
	char *ext = get_ext(file_name);

	
    //if (ext == "png")
	if (strcmp(ext, "png") == 0)
	    type = "png";
	//if (ext == "gif")
	if (strcmp(ext, "gif") == 0)
	    type = "gif";
    //if (ext == "jpg")
	if (strcmp(ext, "jpg") == 0)
        //JPEG is only 24bit.
	    type = "jpeg";

	
	DTImage = NewDTObject(	(APTR)NULL,
							DTA_SourceType, DTST_RAM,
							DTA_BaseName, (IPTR)type,
							PDTA_DestMode, PMODE_V43,
							TAG_DONE);
	if (!DTImage) return(FALSE);

	//Set Color Data	
	CMAP2Object(DTImage, colorBytes, 256*3);

	
	if (GetDTAttrs(DTImage,PDTA_BitMapHeader,(IPTR)&bmhd,TAG_DONE))
	{        
		dtb.MethodID = PDTM_WRITEPIXELARRAY;
		dtb.pbpa_PixelData = bmi->Data;
		dtb.pbpa_PixelFormat = PBPAFMT_LUT8;
		dtb.pbpa_PixelArrayMod = bmi->Width;
		dtb.pbpa_Left = 0;
		dtb.pbpa_Top = 0;
		dtb.pbpa_Width = bmi->Width;
		dtb.pbpa_Height = bmi->Height;

		bmhd->bmh_Width = bmi->Width;
		bmhd->bmh_Height = bmi->Height;
		bmhd->bmh_Depth = 8;
		bmhd->bmh_PageWidth = 320;
		bmhd->bmh_PageHeight = 240;

		DoMethodA(DTImage, (Msg) &dtb);

		//write datatype object to file
		if (file = Open (file_name,MODE_NEWFILE))
		{
			dtw.MethodID = DTM_WRITE;
			dtw.dtw_GInfo = NULL;
			dtw.dtw_FileHandle = file;
			dtw.dtw_Mode = DTWM_RAW;
			dtw.dtw_AttrList = NULL;

			if (DoMethodA(DTImage, (Msg) &dtw)) retval = TRUE;
		}
	}

	if (file) Close (file);
	if (DTImage) DisposeDTObject(DTImage);
	return retval;
}

/**************************************************************************************************/

struct BitMapHeader *readbmhd_by_datatype(char *file_name)
{
	Object *DTImage = NULL;
	struct BitMapHeader *bmh;

	DTImage = NewDTObject(	file_name,
							(DTA_SourceType),	DTST_FILE,
							(DTA_GroupID),		GID_PICTURE,
							(PDTA_Remap),		FALSE,
							(OBP_Precision),	PRECISION_EXACT,
							TAG_DONE);

	if (DTImage)
	{

		if (GetDTAttrs( DTImage,
					PDTA_BitMapHeader, &bmh,
					TAG_END ) == 1)

		{            
            return bmh;
        }
    }
    DisposeDTObject( DTImage );
    return bmh;
}

/**************************************************************************************************/

char *DetectImageType_by_datatype(char *file_name)
{
    char *dtype;
	Object *DTImage = NULL;
	//BitMapHeader *bmh;
	//BOOL test = FALSE;
	struct DataType *dt;

	DTImage = NewDTObject(	file_name,
							(DTA_SourceType),	DTST_FILE,
							(DTA_GroupID),		GID_PICTURE,
							(PDTA_Remap),		FALSE,
							(OBP_Precision),	PRECISION_EXACT,
							TAG_DONE);

	if (DTImage)
	{
        /* Get DataType BaseName. */        
        GetDTAttrs(DTImage, DTA_DataType, (IPTR)&dt, TAG_DONE );        
        dtype = dt->dtn_Header->dth_BaseName;        
    }
    ReleaseDataType(dt);
    DisposeDTObject( DTImage );

    return dtype;
}

/**************************************************************************************************/

BOOL testImage_by_datatype(char *file_name)
{
	Object *DTImage = NULL;
	//BitMapHeader *bmh;
	BOOL test = FALSE;

	DTImage = NewDTObject(	file_name,
							(DTA_SourceType),	DTST_FILE,
							(DTA_GroupID),		GID_PICTURE,
							(PDTA_Remap),		FALSE,
							(OBP_Precision),	PRECISION_EXACT,
							TAG_DONE);

	if (DTImage)
	{
		test = TRUE;
    }
    DisposeDTObject( DTImage );

    return test;
}

/**************************************************************************************************/

/**
 * Sample Code from: CodeReview.StackExchange.com
 * pixels_buffer - Pixels buffer to be modified
 * width - Image width
 * height - Image height
 * bytes_per_pixel - Number of image components, ie: 3 for rgb, 4 rgba, etc...
 **/
void Flip_Vertical(unsigned char *pixels_buffer, LONG width, LONG height, int bytes_per_pixel)
{
    int rowIndex;
    int rows = height / 2; // Iterate only half the buffer to get a full flip
    int row_stride = width * bytes_per_pixel;
    unsigned char* temp_row = (unsigned char*)malloc(row_stride);

    int source_offset, target_offset;
    //** Copy pixel data to temp row then copy last row to first row. Copy temp row back to last row **//
    for (rowIndex = 0; rowIndex < rows; rowIndex++)
    {
        source_offset = rowIndex * row_stride;
        target_offset = (height - rowIndex - 1) * row_stride;

        memcpy(temp_row, pixels_buffer + source_offset, row_stride);
        memcpy(pixels_buffer + source_offset, pixels_buffer + target_offset, row_stride);
        memcpy(pixels_buffer + target_offset, temp_row, row_stride);
    }
    free(temp_row);
    temp_row = NULL;
}

/*********************************************************************************************/

ULONG DoWriteMethod(Object *dto, STRPTR name, struct Window *win, ULONG mode)
{
    //struct Window *win;
    struct dtWrite msg;
    BPTR fh;
    ULONG retval;
    
    fh = BNULL;
    if (name)
    {
        fh = Open( name, MODE_NEWFILE );
        if (!fh)
        {
            //D(bug("Multiview: Cannot open %s\n", name));
            //OutputMessage(MSG(MSG_SAVE_FAILED));
            //return FALSE;
        }
    }    
    
    msg.MethodID          = DTM_WRITE;
    msg.dtw_GInfo         = NULL;
    msg.dtw_FileHandle    = fh;
    msg.dtw_Mode          = mode;
    msg.dtw_AttrList      = NULL;

    //D(bug("Multiview: Saving %s mode %ld\n", name ? name : (STRPTR)"[nothing]", mode));
    retval = DoDTMethodA(dto, win, NULL, (Msg)&msg);
    if (fh)
    {
        Close( fh );
        if( !retval )
        {
            //D(bug("Multiview: Error during write !\n"));
            //OutputMessage(MSG(MSG_SAVE_FAILED));
        }
    }
    return retval;
}

/**************************************************************************************************/

BOOL savepix_by_datatype(unsigned char *image, char *fileToSave, int width, int height, char *type)
{		
	Object *DTImage = NULL;
	struct BitMapHeader *bmhd;
	struct dtWrite dtw;
	struct pdtBlitPixelArray dtb;
	FILE *file = NULL;		
	BOOL retval = FALSE;
	
	//if (type == "jpg")
	if (strcmp(type, "jpg") == 0)
	   type = "jpeg";

	//EasyRequest(NULL, &ResultReq, NULL, type);
	
	DTImage = NewDTObject(	(APTR)NULL,
							DTA_SourceType, DTST_RAM,
							DTA_BaseName, (IPTR)type,
							PDTA_DestMode, PMODE_V43,
							TAG_DONE);
	if (!DTImage) return(FALSE);


	if (GetDTAttrs(DTImage,PDTA_BitMapHeader,(IPTR)&bmhd,TAG_DONE))
	{
		dtb.MethodID = PDTM_WRITEPIXELARRAY;
		dtb.pbpa_PixelData = image;        
		dtb.pbpa_PixelFormat = PBPAFMT_RGB;
		dtb.pbpa_PixelArrayMod = width*3;
		dtb.pbpa_Left = 0;
		dtb.pbpa_Top = 0;
		dtb.pbpa_Width = width;
		dtb.pbpa_Height = height;
		bmhd->bmh_Depth = 24;
		bmhd->bmh_Width = width;
		bmhd->bmh_Height = height;		
		bmhd->bmh_PageWidth = 320;
		bmhd->bmh_PageHeight = 240;

		DoMethodA(DTImage, (Msg) &dtb);

		//write datatype object to file
		if (file = Open (fileToSave,MODE_NEWFILE))
		{			
			dtw.MethodID = DTM_WRITE;
			dtw.dtw_GInfo = NULL;
			dtw.dtw_FileHandle = file;
			dtw.dtw_Mode = DTWM_RAW;
			dtw.dtw_AttrList = NULL;

			if (DoMethodA(DTImage, (Msg) &dtw)) retval = TRUE;
		}
	}

	if (file) Close (file);
	if (DTImage) DisposeDTObject(DTImage);
	
    return retval;
}

/**************************************************************************************************/

BOOL savepix_by_datatype_alpha(unsigned char *image, char *fileToSave, int width, int height, char *type)
{		
	Object *DTImage = NULL;
	struct BitMapHeader *bmhd;
	struct dtWrite dtw;
	struct pdtBlitPixelArray dtb;
	FILE *file = NULL;		
	BOOL retval = FALSE;
	
	//if (type == "jpg")
	if (strcmp(type, "jpg") == 0)
	   type = "jpeg";

	//EasyRequest(NULL, &ResultReq, NULL, type);
	
	DTImage = NewDTObject(	(APTR)NULL,
							DTA_SourceType, DTST_RAM,
							DTA_BaseName, (IPTR)type,
							PDTA_DestMode, PMODE_V43,
							TAG_DONE);
	if (!DTImage) return(FALSE);


	if (GetDTAttrs(DTImage,PDTA_BitMapHeader,(IPTR)&bmhd,TAG_DONE))
	{
		dtb.MethodID = PDTM_WRITEPIXELARRAY;
		dtb.pbpa_PixelData = image;
		dtb.pbpa_PixelFormat = PBPAFMT_ARGB;
		dtb.pbpa_PixelArrayMod = width*4;	        
		dtb.pbpa_Left = 0;
		dtb.pbpa_Top = 0;
		dtb.pbpa_Width = width;
		dtb.pbpa_Height = height;
		bmhd->bmh_Depth = 32;
		bmhd->bmh_Width = width;
		bmhd->bmh_Height = height;		
		bmhd->bmh_PageWidth = 320;
		bmhd->bmh_PageHeight = 240;

		DoMethodA(DTImage, (Msg) &dtb);

		//write datatype object to file
		if (file = Open (fileToSave,MODE_NEWFILE))
		{			
			dtw.MethodID = DTM_WRITE;
			dtw.dtw_GInfo = NULL;
			dtw.dtw_FileHandle = file;
			dtw.dtw_Mode = DTWM_RAW;
			dtw.dtw_AttrList = NULL;

			if (DoMethodA(DTImage, (Msg) &dtw)) retval = TRUE;
		}
	}

	if (file) Close (file);
	if (DTImage) DisposeDTObject(DTImage);
	
    return retval;
}

/**************************************************************************************************/

STRPTR GetFileName(char **filepart, char *titletext)
{
    //static UBYTE pathbuffer[300];
    static UBYTE fileBuffer[300];
    struct FileRequester *req;
    STRPTR retval = NULL;

    AslBase = OpenLibrary("asl.library", 39);
    if (AslBase)
    {
        fileBuffer[299] = 0;
        //pathbuffer[299] = 0;
        *filepart = 0;

        req = AllocAslRequestTags(ASL_FileRequest, ASLFR_TitleText     , (IPTR)titletext,
                                                   ASLFR_InitialLeftEdge, 100,
                                                   ASLFR_InitialTopEdge, 100,
                                                   ASLFR_InitialWidth, 380,
                                                   ASLFR_InitialHeight, 360,
                                                   ASLFR_InitialDrawer, "Sys:",
                                                   ASLFR_InitialPattern, "#?",
                                                   TAG_DONE);
        if (req)        
        {
            if (AslRequest(req, NULL))
            {
                strcpy(fileBuffer, req->fr_Drawer);
                AddPart(fileBuffer, req->fr_File, 299);

                *filepart = req->fr_File;

                retval = fileBuffer;

            } /* if (AslRequest(req, NULL) */

            FreeAslRequest(req);

        } /* if (req) */
        else
        {
            //** User Cancelled **//
            FreeAslRequest(req);
            return retval;
        }

        CloseLibrary(AslBase);

    } /* if (AslBase) */

    return retval;
}

/*********************************************************************************************/

STRPTR GetNewFileName(char **filepart, char *titletext)
{
    //static UBYTE pathbuffer[300];
    static UBYTE filebuffer[300];
    struct FileRequester *req;
    STRPTR retval = NULL;

    AslBase = OpenLibrary("asl.library", 39);
    if (AslBase)
    {
        filebuffer[299] = 0;
        //pathbuffer[299] = 0;
        *filepart = 0;

        req = AllocAslRequestTags(ASL_FileRequest, ASLFR_TitleText     , (IPTR)titletext,
                                                   ASLFR_InitialLeftEdge, 100,
                                                   ASLFR_InitialTopEdge, 100,
                                                   ASLFR_InitialWidth, 380,
                                                   ASLFR_InitialHeight, 360,
                                                   ASLFR_InitialDrawer, "Sys:",
                                                   ASLFR_InitialPattern, "#?",
                                                   TAG_DONE);
        if (req)
        //if (req != NULL)
        {
            if (AslRequest(req, NULL))
            {
                strcpy(filebuffer, req->fr_Drawer);
                AddPart(filebuffer, req->fr_File, 299);

                *filepart = req->fr_File;

                retval = filebuffer;

            } /* if (AslRequest(req, NULL) */

            FreeAslRequest(req);

        } /* if (req) */
        else
        {
            //** User Cancelled **//
            FreeAslRequest(req);
            return retval;
        }

        CloseLibrary(AslBase);

    } /* if (AslBase) */

    return retval;
}

/*********************************************************************************************/
